/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "fe/platform.h"

#define FE_THREAD_DEBUG	FALSE

namespace fe
{

FE_DL_PUBLIC ThreadDefaultInitFunction* Thread::ms_fnDefaultInit=NULL;
FE_DL_PUBLIC ThreadInitFunction* Thread::ms_fnInit=NULL;
FE_DL_PUBLIC ThreadInterruptFunction* Thread::ms_fnInterrupt=NULL;
FE_DL_PUBLIC ThreadJoinFunction* Thread::ms_fnJoin=NULL;
FE_DL_PUBLIC ThreadFinishFunction* Thread::ms_fnFinish=NULL;
FE_DL_PUBLIC ThreadInterruptionFunction* Thread::ms_fnInterruption=NULL;
FE_DL_PUBLIC ThreadConcurrencyFunction* Thread::ms_fnConcurrency=NULL;

FE_DL_PUBLIC ThreadGroupInitFunction* Thread::ms_fnGroupInit=NULL;
FE_DL_PUBLIC ThreadGroupCreateFunction* Thread::ms_fnGroupCreate=NULL;
FE_DL_PUBLIC ThreadGroupJoinAllFunction* Thread::ms_fnGroupJoinAll=NULL;
FE_DL_PUBLIC ThreadGroupFinishFunction* Thread::ms_fnGroupFinish=NULL;

FE_DL_PUBLIC String Thread::ms_support;
FE_DL_PUBLIC BWORD Thread::ms_checked=FALSE;

Thread::Thread(void):
	m_pRawThread(NULL)
{
	confirm();

	if(ms_fnDefaultInit)
	{
		m_pRawThread=ms_fnDefaultInit();
	}

	FEASSERT(m_pRawThread);
}

Thread::Thread(Functor* a_pFunctor):
	m_pRawThread(NULL)
{
	confirm();

	if(ms_fnInit)
	{
		m_pRawThread=ms_fnInit(a_pFunctor);
	}

	FEASSERT(m_pRawThread);
}

Thread::Thread(void* a_pRawThread):
	m_pRawThread(a_pRawThread)
{
	FEASSERT(m_pRawThread);
}

Thread::~Thread(void)
{
	if(m_pRawThread)
	{
		if(ms_fnFinish)
		{
			ms_fnFinish(m_pRawThread);
		}
		m_pRawThread=NULL;
	}
}

BWORD Thread::interrupt(void)
{
	if(!ms_fnInterrupt)
	{
		return FALSE;
	}

	ms_fnInterrupt(m_pRawThread);
	return TRUE;
}

BWORD Thread::join(void)
{
	if(!ms_fnJoin)
	{
		return FALSE;
	}

	ms_fnJoin(m_pRawThread);
	return TRUE;
}

//* static
BWORD Thread::interruptionPoint(void)
{
	if(!ms_fnInterruption)
	{
		return FALSE;
	}

	ms_fnInterruption();
	return TRUE;
}

//* static
I32 Thread::hardwareConcurrency(void)
{
	confirm();

	if(!ms_fnConcurrency)
	{
		return 1;
	}

	return ms_fnConcurrency();
}


Thread::Functor::Functor(void):
	m_pOriginal(this)
{
#if FE_THREAD_DEBUG
	printf("Thread::Functor::Functor %p %p\n",this,m_pOriginal);
#endif
}

Thread::Functor::~Functor(void)
{
#if FE_THREAD_DEBUG
	printf("Thread::Functor::~Functor %p %p\n",this,m_pOriginal);
#endif

	FEASSERT(m_pOriginal);

	m_pOriginal=NULL;
}

void Thread::Functor::operator()(void)
{
#if FE_THREAD_DEBUG
	printf("Thread::Functor::operator() %p %p\n",this,m_pOriginal);
#endif

	FEASSERT(m_pOriginal);

	if(abs(m_pOriginal-this)>0xFFFF)
	{
		printf("Thread::Functor::operator() %p %p CORRUPTED\n",
				this,m_pOriginal);
	}
	else
	{
		if(m_pOriginal!=this)
		{
			printf("Thread::Functor::operator() %p %p CHANGED\n",
					this,m_pOriginal);
		}

		m_pOriginal->operate();
	}

#if FE_THREAD_DEBUG
	printf("Thread::Functor::operator() %p %p DONE\n",this,m_pOriginal);
#endif
}

void Thread::Functor::operate(void)
{
	printf("Thread::Functor::operate doing nothing as base class\n");
}

Thread::Group::Group(void):
	m_pRawThreadGroup(NULL)
{
#if FE_THREAD_DEBUG
	printf("Thread::Group::Group %p raw %p\n",
			this,m_pRawThreadGroup);
#endif

	confirm();

	if(Thread::ms_fnGroupInit)
	{
		m_pRawThreadGroup=Thread::ms_fnGroupInit();
	}

#if FE_THREAD_DEBUG
	printf("Thread::Group::Group %p raw %p DONE\n",
			this,m_pRawThreadGroup);
#endif
}

Thread::Group::~Group(void)
{
#if FE_THREAD_DEBUG
	printf("Thread::Group::~Group %p raw %p\n",
			this,m_pRawThreadGroup);
#endif

	if(Thread::ms_fnGroupFinish)
	{
		Thread::ms_fnGroupFinish(m_pRawThreadGroup);
	}

#if FE_THREAD_DEBUG
	printf("Thread::Group::~Group %p raw %p DONE\n",
			this,m_pRawThreadGroup);
#endif
}

Thread* Thread::Group::createThread(Functor* a_pFunctor)
{
	if(Thread::ms_fnGroupCreate)
	{
#if FE_THREAD_DEBUG
		printf("Thread::Group::createThread %p raw %p functor %p\n",
				this,m_pRawThreadGroup,a_pFunctor);
#endif

		void* pRawThread=Thread::ms_fnGroupCreate(m_pRawThreadGroup,a_pFunctor);

#if FE_THREAD_DEBUG
		printf("Thread::Group::createThread %p raw %p functor %p DONE\n",
				this,m_pRawThreadGroup,a_pFunctor);
#endif

		return new Thread(pRawThread);
	}
	return NULL;
}

BWORD Thread::Group::joinAll(void)
{
	if(Thread::ms_fnGroupJoinAll)
	{
#if FE_THREAD_DEBUG
		printf("Thread::Group::joinAll %p raw %p\n",
				this,m_pRawThreadGroup);
#endif

		Thread::ms_fnGroupJoinAll(m_pRawThreadGroup);

#if FE_THREAD_DEBUG
		printf("Thread::Group::joinAll %p raw %p DONE\n",
				this,m_pRawThreadGroup);
#endif

		return TRUE;
	}

	return FALSE;
}

#ifdef _OPENMP
extern "C"
{
bool omp_thread_init(void);
}
#endif

//* static
BWORD Thread::confirm(void)
{
	if(ms_fnInit)
	{
		return TRUE;
	}
	if(ms_checked)
	{
		return FALSE;
	}
	ms_checked=TRUE;

	const String verbose=System::getVerbose();

	Mutex::confirm();

	//* try using same lib as mutexes
	const String mutexSupport=Mutex::support();
	String nameOrder=mutexSupport;
	if(!nameOrder.empty())
	{
		nameOrder+=":";
	}

	nameOrder+=FE_STRING(FE_THREAD_ORDER);

	DL_Loader *pLoader=new DL_Loader();
	FEASSERT(pLoader);

	BWORD success=FALSE;
	const BWORD adaptname=TRUE;

	String name;
	while(!nameOrder.empty())
	{
		while(!success && !(name=nameOrder.parse("\"",":")).empty())
		{
//			printf("open %s\n",name.c_str());

			if(name=="omp")
			{
#ifdef _OPENMP
				if(verbose=="all")
				{
					printf("  using OpenMP (for threads)\n");
				}
				ms_support=name;

				delete pLoader;
				return omp_thread_init();
#endif
			}

			success=pLoader->openDL(name,adaptname);
		}

		if(!success)
		{
			if(verbose!="none")
			{
				printf("  no thread support\n");
			}
			delete pLoader;
			return FALSE;
		}

		bool (*thread_init)(void);
		thread_init=(bool (*)(void))pLoader->findDLFunction("thread_init");

		if(!thread_init)
		{
			if(verbose!="none")
			{
				fprintf(stderr,"  thread_init not found for %s\n",name.c_str());
			}
		}
		else if(thread_init())
		{
			if(verbose=="all")
			{
				printf("  %s %s (for threading)\n",
						(name==mutexSupport)? "reusing": "using",name.c_str());
			}

			ms_support=name;
			return TRUE;
		}
		else
		{
#if FE_THREAD_DEBUG
			fprintf(stderr,"  thread_init failed for %s\n",name.c_str());
#endif
		}

		delete pLoader;
		success=FALSE;
	}

	return FALSE;

	//* TODO retain, close, and delete pLoader later (leak)
}

} /* namespace */

